#
# Copyright 2008, Martin Owens.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from django.db import models
from django.contrib.auth.models import User
from django.utils.safestring import mark_safe
import re

class PhoneNumber(models.Model):
	areacode  = models.CharField('Area Code', max_length=8, blank=True, null=True)
	prefix    = models.CharField('Prefix', max_length=10, blank=True, null=True)
	number    = models.CharField('Number', max_length=20)
	marketing = models.BooleanField('Marketing Allowed', default=False)
	fax       = models.BooleanField('Fax Number', default=False)

	def __unicode__(self):
		return self.render()

	def render(self, country=None):
		if country:
			return "+0" + str(country) + " " + self.areacode + " " + self.prefix + " " + self.number
		return self.areacode + " " + self.prefix + " " + self.number


class EmailAddress(models.Model):
	address   = models.EmailField('Email Address')
	marketing = models.BooleanField('Marketing Allowed', default=False)
	contact   = models.ForeignKey('Contact', related_name='email_addresses')

	def __unicode__(self):
		return self.address

	class Meta:
		verbose_name_plural = 'email addresses'

# ADDRESS MODELS

class Country(models.Model):
	name       = models.CharField(max_length=255)
	code       = models.CharField(max_length=5)
	phone_code = models.IntegerField()

	def __unicode__(self):
		return self.name

	class Meta:
		verbose_name = 'country'
		verbose_name_plural = 'countries'


class Region(models.Model):
	# AKA state, county
	name    = models.CharField(max_length=255)
	short   = models.CharField(max_length=20)
	country = models.ForeignKey(Country)

	def __unicode__(self):
		return self.name


class Town(models.Model):
	name   = models.CharField(max_length=255)
	region = models.ForeignKey(Region)

	def __unicode__(self):
		return self.name


class Road(models.Model):
	name = models.CharField(max_length=255)
	town = models.ForeignKey(Town)

	def __unicode__(self):
		return self.name


class AddressRender(models.Model):
	template = models.TextField()
	country  = models.ForeignKey(Country)

	def __unicode__(self):
		return "Template for %s" % unicode(self.country)


class MobilePhone(PhoneNumber):
	contact = models.ForeignKey('Person')


class AddressPhone(PhoneNumber):
	address = models.ForeignKey('Address')


class Address(models.Model):
	room     = models.CharField('Apartment/Room Number', max_length=100, blank=True, null=True)
	building = models.CharField('Building Name', max_length=100, blank=True, null=True)
	number   = models.CharField('House Number', max_length=10, blank=True, null=True)
	road     = models.ForeignKey(Road, blank=True, null=True)
	postcode = models.CharField('Zipcode', max_length=20, blank=True, null=True)
	marketing = models.BooleanField('Marketing Allowed', default=False)

	def __unicode__(self):
		return self.render()

	def address(self):
		return self.render()

	def phone(self):
		if self.phone:
			return self.phone.render( self.country().phone_code )
		return ''

	def town(self):
		if self.road:
			return self.road.town
		return None

	def region(self):
		if self.town():
			return self.town().region
		return None

	def country(self):
		if self.region():
			return self.region().country
		return None

	def render(self, name='', template=None, html=False):
		"""
		This method should attempt to create a mailing label.
		"""

		if not template:
			#template = AddressRender.get(country=self.country())
			template = None

			if template:
				template = unicode(template)
			else:
				template = """[%(name)s
][%(building)s
][%(number)s [%(road)s][, #%(room)s]
][%(town)s, ][%(sregion)s ][%(postcode)s
][%(country)s]"""

		items = {
			'name'      : name,
			'room'      : self.room,
			'building'  : self.building,
			'number'    : self.number,
			'postcode'  : self.postcode,
			'road':'','town':'','region':'','sregion':'','country':'',
		}
		if self.road:
			items['road']    = unicode(self.road)
			items['town']    = unicode(self.road.town)
			items['region']  = unicode(self.road.town.region)
			items['sregion'] = unicode(self.road.town.region.short)
			items['country'] = unicode(self.country())

		result = template % items
		result = re.sub(r'\[[\.,\ \:\'\"\;\!\@\#\$\%\^\&\*\n\r\t]*\]', '', result).replace('[','').replace(']','')
		if html:
			result = mark_safe(result.replace('\n', '<br/>'))
		return result


	class Meta:
		verbose_name = 'address'
		verbose_name_plural = 'addresses'


# CONTACTS

class PersonType(models.Model):
	name = models.CharField(max_length=60)

	def __unicode__(self):
		return self.name


class OrganisationType(models.Model):
	name = models.CharField(max_length=60)

	def __unicode__(self):
		return self.name


class Contact(models.Model):
	website  = models.URLField('Website', null=True, blank=True)
	notes    = models.TextField('Notes', null=True, blank=True)
	created  = models.DateTimeField('Date Created', auto_now_add=True)
	edited   = models.DateTimeField('Date Modified', auto_now=True)

	def __unicode__(self):
		return "Contact %d" % self.id


class Person(Contact):
	type         = models.ForeignKey(PersonType)
	site_user    = models.ForeignKey(User, null=True, blank=True)
	title        = models.CharField('Title', max_length=5, null=True, blank=True)
	first_name   = models.CharField('First Name', max_length=100)
	middle_name  = models.CharField('Middle Name', max_length=100, null=True, blank=True)
	last_name    = models.CharField('Last Name', max_length=100, null=True, blank=True)
	primary_address = models.ForeignKey(Address, null=True, blank=True, related_name='residents')
	other_address   = models.ManyToManyField(Address, related_name='other_person', null=True, blank=True)
	organisation = models.ForeignKey('Organisation', blank=True, null=True)

	def __unicode__(self):
		"""
		Puts together the best name for a person from the component parts.
		"""
		return self.first_name + " " + self.last_name

	class Meta:
		verbose_name = 'person'
		verbose_name_plural = 'people'


class Organisation(Contact):
	type    = models.ForeignKey(OrganisationType)
	name    = models.CharField('Organisation Name', max_length=255)
	contact = models.ForeignKey(Person, verbose_name='Contact Person', related_name='contact_of', null=True, blank=True)

	def __unicode__(self):
		return self.name
