#
# Copyright 2008, Martin Owens.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from django.db import models
import moapps.contacts.models as contact

PAYMENTS = (
	( 'CASH', 'Cash' ),
	( 'CHEQ', 'Cheque' ),
	( 'CRED', 'Credit Card' ),
)

PAYSTATUS = (
	( 'PEND', 'Pending' ),
	( 'DONE', 'Complete' ),
	( 'ERR',  'Failed' ),
)

SEAT_TYPES = (
	( 'AB', 'Section A, Balcony' ),
	( 'AF', 'Section A, Floor' ),
	( 'B',  'Section B' ),
	( 'C',  'Section C' ),
	( 'D',  'Section D' ),
)

TICKET_TYPES = (
	( 'SINGLE', 'Single Ticket' ),
	( 'SEASON',  'Subscription' ),
)

def g( t, n ):
	for x in t:
		if x[0] == n:
			return x[1]
	return n


class Season(models.Model):
	name        = models.CharField('Season Name', max_length=100)
	description = models.TextField(null=True, blank=True)

	def __unicode__(self):
		return self.name


class SeasonPerformance(models.Model):
	season = models.ForeignKey(Season)
	date   = models.DateTimeField('Performance Date')
	name   = models.CharField('Performance Name', max_length=100, blank=True, null=True)

	def __unicode__(self):
		if self.name:
			return self.name
		else:
			return str(self.season) + " on the " + str(self.date.strftime("%a %d %b %Y"))


class TicketProduct(models.Model):
	sku         = models.CharField('Ticket Code', max_length=100)
	seat_type   = models.CharField('Section Type', max_length=2, choices=SEAT_TYPES)
	ticket_type = models.CharField('Ticket Type', max_length=10, choices=TICKET_TYPES)
	cost_each   = models.FloatField('Cost')
	stock_level = models.IntegerField('Number Available')

	def __unicode__(self):
		return "%s: %s, %s ($%0.2f)" % (self.sku, g(SEAT_TYPES,self.seat_type), g(TICKET_TYPES,self.ticket_type), self.cost_each)


class OrderStatus(models.Model):
	name = models.CharField(max_length=50)

	def __unicode__(self):
		return self.name

	class Meta:
		verbose_name = 'Order Status'
		verbose_name_plural = 'Order Status'


class Order(models.Model):
	orderer = models.ForeignKey(contact.Person)
	ordered = models.DateTimeField("Ordered Date", auto_now_add=True)
	shipped = models.DateTimeField("Shipped Date", blank=True, null=True)
	status  = models.ForeignKey(OrderStatus)
	address = models.ForeignKey(contact.Address, related_name='shipping_address', null=True, blank=True)
	notes   = models.TextField(blank=True, null=True)
	handle  = models.FloatField("Handling Fee", default=5)

	def __unicode__(self):
		return "Order %d by %s on %s" % (self.id, str(self.orderer), str(self.ordered))

#	def total(self):
#		items = self.ticketorderitem.all()


class Payment(models.Model):
	order   = models.ForeignKey(Order)
	type    = models.CharField(max_length=10, choices=PAYMENTS)
	status  = models.CharField(max_length=10, choices=PAYSTATUS)
	address = models.ForeignKey(contact.Address, related_name='payment_address')

	def __unicode__(self):
		return "%s (%s)" % (g(PAYMENTS, self.type), g(PAYSTATUS, self.status))


class TicketOrderItem(models.Model):
	order       = models.ForeignKey(Order)
	cost        = models.FloatField('Cost Each')
	quantity    = models.IntegerField('Quantity', default=1)
	product     = models.ForeignKey(TicketProduct)
	performance = models.ManyToManyField(SeasonPerformance, null=True, blank=True)

	def __unicode__(self):
		p = []
		for perf in self.performance.all():
			p.append( str( perf ) ) 
		return "%s costing $%.2f" % (str(self.product), self.cost)


class DonationReason(models.Model):
	reason = models.CharField(max_length=100)
	
	def __unicode__(self):
		return self.reason


class DonationOrderItem(models.Model):
	order  = models.ForeignKey(Order)
	cost   = models.FloatField('Donation Amount')
	reason = models.ForeignKey(DonationReason, blank=True, null=True)
	
	def __unicode__(self):
		return "%s Donation of $%.2f" % (str(self.order), self.cost)

