#
# Copyright 2008, Martin Owens.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys, os
from forms import BasicIdForm
from django import template
from django.utils.safestring import mark_safe
from django.forms.models import modelform_factory
from django.db.models.base import ModelBase

blocks = {}


def get_plural_label(model):
	label = model._meta.verbose_name_plural
	if type(label) not in [str, unicode]:
		label = model._meta.verbose_name+'s'
	return label.title()


class BlockBase:
	def getId(self):
		return self.getType().lower()

	def getType(self):
		return self.getModel().__name__


class ModelBlockBase(BlockBase):
	def __init__(self, model):
		self.model = model

	def getModel(self):
		return self.model


class ObjectBlockBase(BlockBase):
	def __init__(self, object, parents=[]):
		self.object  = object
		self.parents = parents
	
	def getModel(self):
		return type(self.object)

	def getFields(self):
		return self.object._meta.fields

	def getRelated(self):
		return self.object._meta.get_all_related_objects()


class IndexBlock(ModelBlockBase):
	def list(self):
		return self.model._default_manager.all()


class ViewBlock(ObjectBlockBase):
	def view_template(self):
		return 'workflows/data_view.html'

	def content_template(self):
		return 'workflows/data_content.html'

	def list_template(self):
		return self.view_template()

	def list_content_template(self):
		return self.content_template()

	def view(self, view_label=''):
		"""
		Return an html safe string containing the title followed by the contents.
		"""
		if view_label=='':
			view_label = self.getType().title()
		view_template = template.loader.get_template(self.view_template())
		data  = template.Context({'content': self.content(), 'object': self.object, 'label': view_label})
		return mark_safe(view_template.render(data))


	def content(self):
		"""
		Return an html safe string to go below the section title.
		"""
		view_template = template.loader.get_template(self.content_template())
		data  = template.Context({'data': self.data(), 'object': self.object})
		return mark_safe(view_template.render(data))


	def list_view(self, list, label, no_block=False):
		"""
		Return an html safe string which contains a list of related objects.
		"""
		view_template = template.loader.get_template(self.list_template())
		data = template.Context({'content': mark_safe(self.list_content(list, no_block)), 'label': label})
		return mark_safe(view_template.render(data))


	def list_content(self, list, no_block=False):
		item_template = template.loader.get_template(self.list_content_template())
		item_data = []
		for item in list:
			datum  = None
			blocks = getBlocks( item )
			if blocks:
				if no_block:
					datum = self.create_link( item )
				else:
					datum = self.block_datum( blocks.id, item, None )
			item_data.append( { 'label': None, 'value': datum or str(item) } )
		return item_template.render( template.Context({'data':item_data}) )


	def data(self):
		"""
		Return a standard set of data fields for viewing in a normal table.
		"""
		return self.field_data() + self.related_data()


	def field_data(self):
		data = []
		for field in self.getFields():
			if field.name not in self.ignored_fields():
				data.append( self.datum( field.name ) )
		return data


	def related_data(self):
		data = []
		for related in self.getRelated():
			name = related.get_accessor_name()
			if name not in self.ignored_relations():
				list = getattr(self.object, name).all()
				if list:
					id    = type(list[0]).__name__.lower()
					label  = name.replace('_',' ').title()
					if '_set' in name:
						label = get_plural_label(type(list[0]))
					no_block = name in self.simple_relations()
					sys.stderr.write("Looking at Relation '%s'\n" % name)
					data.append( { 'label': None, 'value': self.list_view( list, label.title(), no_block ) } )
		return data


	def datum(self, name, datum=None):
		"""
		Return a single datum entry based in the field name.
		"""
		if datum==None:
			datum = getattr(self.object, name)
		if (not name or name != 'id' and name[-4:] != '_ptr') and datum and datum not in self.parents:
			label = name.replace('_', ' ').title()
			subtype = type(datum).__name__.lower()
			if getBlocks( subtype ):
				sys.stderr.write("Looking at Field '%s', %s\n" % (name, str(datum)) )
				if name not in self.simple_fields():
					result = self.block_datum( subtype, datum, label )
					if result:
						return { 'label':None, 'value':result }
				datum = self.create_link( datum )
			return { 'label': label, 'value': datum }


	def create_link(self, object):
		return mark_safe( "<a href='../../%s/%d/'>%s</a>" % (getBlockId(object), object.id, str(object)) )


	def block_datum(self, subtype, datum, label=None):
		parents = self.parents + [ self.object ]
		return getBlocks( subtype ).view( datum, parents ).view( view_label=label )


	def ignored_fields(self):
		return []
	def ignored_relations(self):
		return []
	def simple_fields(self):
		return []
	def simple_relations(self):
		return []
	


class EditBlock(ObjectBlockBase):
	def form(self):
		# OK, so here we need to ignore any forien key based fields
		# This is so we can impliment a full set of fields to create
		# A new object. This is ultimatly how a form gets constructed
		# And how the relationships work out.
		# form=ModelForm, fields=None, exclude=None, formfield_callback=lambda f: f.formfield()
		return modelform_factory( self.getModel() )

	def getForm(self, *opts, **attrs ):
		# Each form should have prefix so the data isn't confused.
		attrs['instance'] = self.object
		return self.form()( *opts, **attrs )

	def save(self, data):
		# Here we're not only taking the current form but we need to take all
		# The subForms, loop through them all, create each of the items required
		# and pass back the values.
		pass

	def subForms(self):
		# Return a list of forms which allow the user to create new data
		pass


class CreateBlock(ModelBlockBase):
	pass


class DiscoveryBlock(ModelBlockBase):
	def getForm(self, *opts, **attrs):
		return self.form()(*opts, **attrs)

	def form(self):
		return BasicIdForm

	def return_field(self):
		return 'id'


class Blocks:
	"""
	Class for holding each of the block classes for a given model.
	"""
	def __init__(self, model, view=None, edit=None, create=None, discovery=None, index=None, indexed=True ):
		self.model     = model
		self.view      = view or ViewBlock
		self.edit      = edit or EditBlock
		self.create    = create and create(self.model) or CreateBlock(self.model)
		self.discovery = discovery and discovery(self.model) or DiscoveryBlock(self.model)
		self.index     = index and index(self.model) or IndexBlock(self.model)
		self.indexed   = indexed
		self.id        = self.index.getId()

	def __unicode__(self):
		return get_plural_label(self.model)



def register_blocks( model, id=None, **attrs ):
	"""
	Register a new set of blocks, same inputs as the Blocks class init.
	"""
	if model and type(model) != str:
		if not id:
			id = model.__name__.lower()
		if blocks.has_key( id ):
			sys.stderr.write("Blocks for %s model already registered!\n" % id)
		blocks[id] = Blocks( model, **attrs )


def getBlocks( o ):
	"""
	Return a set of blocks based on an id.
	"""
	id = getBlockId( o )
	sys.stderr.write("found id %s from %s\n" % (id, str(o)))
	if blocks.has_key( id ):
		return blocks[id]
	return None


def getBlockId( source ):
	"""
	Return a block id based on a model, object or string.
	"""
	if type(source) in [ str, unicode ]:
		return source
	elif type(source)==ModelBase:
		return source.__name__.lower()
	return type(source).__name__.lower()

